/*
** Module   : com.goldencode.p2j.xml
** Abstract : SAX reader handler specific features implementation
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVK 20131030 Created initial version.
*/
package com.goldencode.p2j.xml;

import com.goldencode.p2j.util.*;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import java.io.IOException;
import java.util.*;


class ProcDefaultHandler extends DefaultHandler
{
   /** The procedure name to be associated with the given handle object.*/
   private final handle procHandle;

   /** The SaxReader instance to be associated with the given handle object.*/
   private final handle saxReader;

   /**
    * Default constructor accepting a handle to procedure and SaxReader instance.
    *
    * @param    procHandle
    *           The procedure name to be associated with the given handle object.
    * @param    saxReader
    *           The SaxReader instance to be associated with the given handle object.
    */
   public ProcDefaultHandler(handle procHandle, handle saxReader)
   {
      this.procHandle = procHandle;
      this.saxReader = saxReader;
   }

   /**
    * Resolve an external entity.
    *
    * <p>Always return null, so that the parser will use the system
    * identifier provided in the XML document.  This method implements
    * the SAX default behaviour: application writers can override it
    * in a subclass to do special translations such as catalog lookups
    * or URI redirection.</p>
    *
    * @param publicId The public identifier, or null if none is
    *                 available.
    * @param systemId The system identifier provided in the XML
    *                 document.
    * @return The new input source, or null to require the
    *         default behaviour.
    * @exception java.io.IOException If there is an error setting
    *            up the new input source.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.EntityResolver#resolveEntity
    */
   @Override
   public InputSource resolveEntity(String publicId, String systemId)
           throws IOException, SAXException
   {
      character filePath = new character();
      longchar memPointer = new longchar(); //todo check memptr
      invokeProcedure(CallBack.ResolveEntity,
                      new character(publicId),
                      filePath,
                      memPointer);
      if (!filePath.isUnknown() || !memPointer.isUnknown())
      {
         //todo throw error
      }

      InputSource is = null;
      if (!filePath.isUnknown())
      {
         String file = filePath.toStringMessage();
         is = new InputSource(new SaxReaderImpl.FileInputStreamWrapper(file));
      }
      else if (!memPointer.isUnknown())
      {
         is = new InputSource(new SaxReaderImpl.LongcharInputStream(memPointer));
      }

      return is;
   }

   /**
    * Receive notification of a notation declaration.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass if they wish to keep track of the notations
    * declared in a document.</p>
    *
    * @param name The notation name.
    * @param publicId The notation public identifier, or null if not
    *                 available.
    * @param systemId The notation system identifier.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.DTDHandler#notationDecl
    */
   @Override
   public void notationDecl(String name, String publicId, String systemId) throws SAXException
   {
      invokeProcedure(CallBack.NotationDecl,
                      new character(name),
                      new character(publicId),
                      new character(systemId));
   }

   /**
    * Receive notification of an unparsed entity declaration.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass to keep track of the unparsed entities
    * declared in a document.</p>
    *
    * @param name The entity name.
    * @param publicId The entity public identifier, or null if not
    *                 available.
    * @param systemId The entity system identifier.
    * @param notationName The name of the associated notation.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.DTDHandler#unparsedEntityDecl
    */
   @Override
   public void unparsedEntityDecl(String name, String publicId, String systemId,
                                  String notationName) throws SAXException
   {
      invokeProcedure(CallBack.UnparsedEntityDecl,
                      new character(name),
                      new character(publicId),
                      new character(systemId),
                      new character(notationName));
   }

   /**
    * Receive a Locator object for document events.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass if they wish to store the locator for use
    * with other document events.</p>
    *
    * @param locator A locator for all SAX document events.
    * @see org.xml.sax.ContentHandler#setDocumentLocator
    * @see org.xml.sax.Locator
    */
   @Override
   public void setDocumentLocator(Locator locator)
   {
      super.setDocumentLocator(locator);    //TODO autogenerated code by evk, please fix this
   }

   /**
    * Receive notification of the beginning of the document.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the beginning
    * of a document (such as allocating the root node of a tree or
    * creating an output file).</p>
    *
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#startDocument
    */
   @Override
   public void startDocument() throws SAXException
   {
      invokeProcedure(CallBack.StartDocument);
   }

   /**
    * Receive notification of the end of the document.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the end
    * of a document (such as finalising a tree or closing an output
    * file).</p>
    *
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#endDocument
    */
   @Override

   public void endDocument() throws SAXException
   {
      invokeProcedure(CallBack.EndDocument);
   }

   /**
    * Receive notification of the start of a Namespace mapping.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the start of
    * each Namespace prefix scope (such as storing the prefix mapping).</p>
    *
    * @param prefix The Namespace prefix being declared.
    * @param uri The Namespace URI mapped to the prefix.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#startPrefixMapping
    */
   @Override
   public void startPrefixMapping(String prefix, String uri) throws SAXException
   {
      invokeProcedure(CallBack.StartPrefixMapping, new character(prefix), new character(uri));
   }

   /**
    * Receive notification of the end of a Namespace mapping.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the end of
    * each prefix mapping.</p>
    *
    * @param prefix The Namespace prefix being declared.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#endPrefixMapping
    */
   @Override
   public void endPrefixMapping(String prefix) throws SAXException
   {
      invokeProcedure(CallBack.EndPrefixMapping, new character(prefix));
   }

   /**
    * Receive notification of the start of an element.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the start of
    * each element (such as allocating a new tree node or writing
    * output to a file).</p>
    *
    * @param uri The Namespace URI, or the empty string if the
    *        element has no Namespace URI or if Namespace
    *        processing is not being performed.
    * @param localName The local name (without prefix), or the
    *        empty string if Namespace processing is not being
    *        performed.
    * @param qName The qualified name (with prefix), or the
    *        empty string if qualified names are not available.
    * @param attributes The attributes attached to the element.  If
    *        there are no attributes, it shall be an empty
    *        Attributes object.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#startElement
    */
   @Override
   public void startElement(String uri, String localName, String qName, Attributes attributes)
           throws SAXException
   {
      /*
      * Callback error: SAX-PARSE sent procedure StartElement mismatched parameters.(10512)
      * */

      invokeProcedure(CallBack.StartElement,
                       new character(uri),
                       new character(localName),
                       new character(qName),
                       new handle(new SaxAttributesImpl(attributes)));

   }

   /**
    * Receive notification of the end of an element.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions at the end of
    * each element (such as finalising a tree node or writing
    * output to a file).</p>
    *
    * @param uri The Namespace URI, or the empty string if the
    *        element has no Namespace URI or if Namespace
    *        processing is not being performed.
    * @param localName The local name (without prefix), or the
    *        empty string if Namespace processing is not being
    *        performed.
    * @param qName The qualified name (with prefix), or the
    *        empty string if qualified names are not available.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#endElement
    */
   @Override
   public void endElement(String uri, String localName, String qName) throws SAXException
   {
      invokeProcedure(CallBack.EndElement, new character(uri), new character(localName),
                      new character(qName));
   }

   /**
    * Receive notification of character data inside an element.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method to take specific actions for each chunk of character data
    * (such as adding the data to a node or buffer, or printing it to
    * a file).</p>
    *
    * @param ch The characters.
    * @param start The start position in the character array.
    * @param length The number of characters to use from the
    *               character array.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#characters
    */
   @Override
   public void characters(char[] ch, int start, int length) throws SAXException
   {
      invokeProcedure(CallBack.Characters, new longchar(new String(ch, start, length)), length);
   }

   /**
    * Receive notification of ignorable whitespace in element content.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method to take specific actions for each chunk of ignorable
    * whitespace (such as adding data to a node or buffer, or printing
    * it to a file).</p>
    *
    * @param ch The whitespace characters.
    * @param start The start position in the character array.
    * @param length The number of characters to use from the
    *               character array.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#ignorableWhitespace
    */
   @Override
   public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
   {
      invokeProcedure(CallBack.IgnorableWhitespace, new longchar(new String(ch, start, length)),
                      new integer(length));
   }

   /**
    * Receive notification of a processing instruction.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions for each
    * processing instruction, such as setting status variables or
    * invoking other methods.</p>
    *
    * @param target The processing instruction target.
    * @param data The processing instruction data, or null if
    *             none is supplied.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#processingInstruction
    */
   @Override
   public void processingInstruction(String target, String data) throws SAXException
   {
      invokeProcedure(CallBack.ProcessingInstruction, new character(target), new character(data));
   }

   /**
    * Receive notification of a skipped entity.
    *
    * <p>By default, do nothing.  Application writers may override this
    * method in a subclass to take specific actions for each
    * processing instruction, such as setting status variables or
    * invoking other methods.</p>
    *
    * @param name The name of the skipped entity.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ContentHandler#processingInstruction
    */
   @Override
   public void skippedEntity(String name) throws SAXException
   {
   }

   /**
    * Receive notification of a parser warning.
    *
    * <p>The default implementation does nothing.  Application writers
    * may override this method in a subclass to take specific actions
    * for each warning, such as inserting the message in a log file or
    * printing it to the console.</p>
    *
    * @param e The warning information encoded as an exception.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ErrorHandler#warning
    * @see org.xml.sax.SAXParseException
    */
   @Override
   public void warning(SAXParseException e) throws SAXException
   {
      invokeProcedure(CallBack.Warning, new character(e.getMessage()));
   }

   /**
    * Receive notification of a recoverable parser error.
    *
    * <p>The default implementation does nothing.  Application writers
    * may override this method in a subclass to take specific actions
    * for each error, such as inserting the message in a log file or
    * printing it to the console.</p>
    *
    * @param e The error information encoded as an exception.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ErrorHandler#warning
    * @see org.xml.sax.SAXParseException
    */
   @Override
   public void error(SAXParseException e) throws SAXException
   {
      invokeProcedure(CallBack.Error, new character(e.getMessage()));
   }

   /**
    * Report a fatal XML parsing error.
    *
    * <p>The default implementation throws a SAXParseException.
    * Application writers may override this method in a subclass if
    * they need to take specific actions for each fatal error (such as
    * collecting all of the errors into a single report): in any case,
    * the application must stop all regular processing when this
    * method is invoked, since the document is no longer reliable, and
    * the parser may no longer report parsing events.</p>
    *
    * @param e The error information encoded as an exception.
    * @exception org.xml.sax.SAXException Any SAX exception, possibly
    *            wrapping another exception.
    * @see org.xml.sax.ErrorHandler#fatalError
    * @see org.xml.sax.SAXParseException
    */
   @Override
   public void fatalError(SAXParseException e) throws SAXException
   {
      invokeProcedure(CallBack.FatalError, new character(e.getMessage()));
   }


   /**
    * Method can invoke procedure callback and valid input parameters. If specified callback is
    * not exists then do nothing.
    *
    * @param    callBack
    *           Callback entity.
    * @param    args
    *           Array of input/output arguments.
    */
   private void invokeProcedure(CallBack callBack, Object... args) throws SAXException
   {
      if (!procHandle._isValid())
      {

      }
      ExternalProgramWrapper programWrapper = ((ExternalProgramWrapper)procHandle.getResource());
      character entries = programWrapper.internalEntries();
      if (!entries.isUnknown())
      {
         if (!entries.toStringMessage().toUpperCase().contains(callBack.name().toUpperCase()))
         {
            return;
         }
      }
      String inputParts = programWrapper.getSignature(callBack.name()).toStringMessage();
      List<String> inputParameters = new ArrayList<>();
      List<String> outputParameters = new ArrayList<>();
      String[] inputPartsArr = inputParts.split(",");
      for (String par : inputPartsArr)
      {
         par = par.trim();
         if (par.startsWith("INPUT")) //input parameter
         {
            inputParameters.add(par.substring(par.lastIndexOf(' ') + 1).toLowerCase());
         }
         else if (par.startsWith("OUTPUT")) //output parameter
         {
            outputParameters.add(par.substring(par.lastIndexOf(' ') + 1).toLowerCase());
         }
      }

      for (int i = 0; i < inputParameters.size(); i++)
      {
         if (!callBack.isValidInputParameterType(inputParameters.get(i), i))
         {
            throw new MismatchedParameters(callBack.name());
         }
      }

      for (int i = 0; i < outputParameters.size(); i++)
      {
         if (!callBack.isValidOutputParameterType(outputParameters.get(i), i))
         {
            throw new MismatchedParameters(callBack.name());
         }
      }

      SelfManager.pushSelf(saxReader);
      try
      {
         ControlFlowOps.invokeInWithMode(callBack.name(), procHandle, callBack.getModes(), args);
      }
      catch (Exception ignored) // just ignore exception
      {
      }
      finally
      {
         SelfManager.popSelf();
      }
   }

   /**
    * List of possible callbacks of SAX-READER object:
    * <ul>
    * <li>{@link #StartDocument} -Start Document callback.</li>
    * <li>{@link #EndDocument} - End Document callback.</li>
    * <li>{@link #ResolveEntity} - ResolveEntity callback.</li>
    * <li>{@link #NotationDecl} - Notation Declaration callback.</li>
    * <li>{@link #UnparsedEntityDecl} - Unparsed Entity Declaration callback.</li>
    * <li>{@link #StartPrefixMapping} - Start Prefix Mapping callback.</li>
    * <li>{@link #EndPrefixMapping} - End Prefix Mapping callback.</li>
    * <li>{@link #StartElement} - Start Element callback.</li>
    * <li>{@link #EndElement} - End Element callback.</li>
    * <li>{@link #Characters} - Characters callback.</li>
    * <li>{@link #IgnorableWhitespace} - Ignorable Whitespace callback.</li>
    * <li>{@link #ProcessingInstruction} - Processing Instruction callback.</li>
    * <li>{@link #Warning} - Warning callback.</li>
    * <li>{@link #Error} - Error callback.</li>
    * <li>{@link #FatalError} - FatalError callback.</li>
    * </ul>
    */
   enum CallBack
   {
      StartDocument(),
      EndDocument(),
      ResolveEntity(new String[]{"character", "character"}, new String[]{"character", "longchar"},
                    "IIOO"),  //maybe memptr
      NotationDecl(new String[]{"character", "character", "character"}, "III"),
      UnparsedEntityDecl(new String[]{"character", "character", "character", "character"}, "IIII"),
      StartPrefixMapping(new String[]{"character", "character"}, "II"),
      EndPrefixMapping(new String[]{"character"}, "I"),
      StartElement(new String[]{"character", "character", "character", "handle"}, "IIII"),
      EndElement(new String[]{"character", "character", "character"}, "III"),
      Characters(new String[]{"longchar","integer"}, "II"),//todo maybe memptr
      IgnorableWhitespace(new String[]{"longchar", "integer"}, "II"),
      ProcessingInstruction(new String[]{"character", "character"}, "II"),
      Warning(new String[]{"character"}, "I"),
      Error(new String[]{"character"}, "I"),
      FatalError(new String[]{"character"}, "I");

      /** possible input parameters types*/
      private final String[] inputArgs;
      /** possible output parameters types*/
      private final String[] outputArgs;
      /** defined input/output modes for procedure*/
      private final String modes;

      /** default constructor*/
      CallBack()
      {
         this(new String[0], new String[0], "");
      }

      /**
       * Special constructor accepting array with possible input type parameters.
       *
       * @param    inputArgs
       *           String array of possible input parameters name.
       */
      CallBack(String[] inputArgs)
      {
         this(inputArgs, new String[0], "");
      }

      /**
       * Special constructor accepting array with possible input/output type parameters.
       *
       * @param    inputArgs
       *           String array of possible input parameters name.
       * @param    outputArgs
       *           String array of possible output parameters name.
       */
      CallBack(String[] inputArgs, String[] outputArgs)
      {
         this(inputArgs, outputArgs, "");
      }

      /**
       * Special constructor accepting array with possible input type parameters.
       *
       * @param    inputArgs
       *           String array of possible input parameters name.
       * @param    modes
       *           String of possible input/output parameters type, e.g "I", "O", "U".
       */
      CallBack(String[] inputArgs, String modes)
      {
         this(inputArgs, new String[0], modes);
      }

      /**
       * Special constructor accepting array with possible input type parameters.
       *
       * @param    inputArgs
       *           String array of possible input parameters name.
       * @param    outputArgs
       *           String array of possible output parameters name.
       * @param    modes
       *           String of possible input/output parameters type, e.g "I", "O", "U".
       */
      CallBack(String[] inputArgs, String[] outputArgs, String modes)
      {
         this.inputArgs = inputArgs;
         this.outputArgs = outputArgs;
         this.modes = modes;
      }

      /**
       * Gets the input/output parameter modes.
       *
       * @return   see above.
       */
      String getModes()
      {
         return modes;
      }

      /**
       * Validate input parameter type with specific index.
       *
       * @param    type
       *           String name of input type.
       * @param    index
       *           index of type in procedure.
       *
       * @return   true if type name is correct for specific position.
       */
      boolean isValidInputParameterType(String type, int index)
      {
         return isValidParameterType(inputArgs, type, index);
      }

      /**
       * Validate output parameter type with specific index.
       *
       * @param    type
       *           String name of input type.
       * @param    index
       *           index of type in procedure.
       *
       * @return   true if type name is correct for specific position.
       */
      boolean isValidOutputParameterType(String type, int index)
      {
         return isValidParameterType(outputArgs, type, index);
      }

      /**
       * Validate intput/output parameter type with specific index.
       *
       * @param    args
       *           String array of possible name types.
       * @param    type
       *           String name of input type.
       * @param    index
       *           index of type in procedure.
       *
       * @return   true if type name is correct for specific position.
       */
      private boolean isValidParameterType(String[] args, String type, int index)
      {
         if (index < 0 || args.length <= index)
         {
            return false;
         }
         return args[index].equals(type);
      }

   }

   /**
    * Special error throws if some parameter mismatch type.
    */
   static class MismatchedParameters extends SAXException
   {
      /** name of callback*/
      String callbackName;

      /**
       * Create a new MismatchedParameters.
       *
       * @param callbackName The callback name.
       */
      MismatchedParameters(String callbackName)
      {
         this.callbackName = callbackName;
      }
   }
}
